"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.Where = exports.Op = void 0;
const BindParam_1 = require("../BindParam/BindParam");
const __1 = require("../..");
const Errors_1 = require("../../Errors");
const Literal_1 = require("../Literal");
/** symbols for Where operations */
const OpEq = Symbol('eq');
const OpIn = Symbol('in');
const Op_In = Symbol('_in');
const OpContains = Symbol('contains');
const OpGt = Symbol('gt');
const OpGte = Symbol('gte');
const OpLt = Symbol('lt');
const OpLte = Symbol('lte');
const OpNe = Symbol('ne');
exports.Op = {
    eq: OpEq,
    in: OpIn,
    _in: Op_In,
    contains: OpContains,
    gt: OpGt,
    gte: OpGte,
    lt: OpLt,
    lte: OpLte,
    ne: OpNe,
};
const operators = [
    'eq',
    'in',
    '_in',
    'contains',
    'gt',
    'gte',
    'lt',
    'lte',
    'ne',
];
const isOperator = {
    eq: (value) => typeof value === 'object' && value && exports.Op.eq in value,
    in: (value) => typeof value === 'object' && value && exports.Op.in in value,
    _in: (value) => typeof value === 'object' && value && exports.Op._in in value,
    contains: (value) => typeof value === 'object' && value && exports.Op.contains in value,
    gt: (value) => typeof value === 'object' && value && exports.Op.gt in value,
    gte: (value) => typeof value === 'object' && value && exports.Op.gte in value,
    lt: (value) => typeof value === 'object' && value && exports.Op.lt in value,
    lte: (value) => typeof value === 'object' && value && exports.Op.lte in value,
    ne: (value) => typeof value === 'object' && value && exports.Op.ne in value,
};
const isNeo4jSupportedTypes = (value) => {
    const isSupportedSingleType = (value) => {
        return (value instanceof Literal_1.Literal ||
            typeof value === 'string' ||
            typeof value === 'number' ||
            typeof value === 'boolean' ||
            __1.neo4jDriver.isInt(value) ||
            __1.neo4jDriver.isPoint(value) ||
            __1.neo4jDriver.isDate(value) ||
            __1.neo4jDriver.isTime(value) ||
            __1.neo4jDriver.isLocalTime(value) ||
            __1.neo4jDriver.isDateTime(value) ||
            __1.neo4jDriver.isLocalDateTime(value) ||
            __1.neo4jDriver.isDuration(value));
    };
    if (Array.isArray(value)) {
        return value.every((element) => isSupportedSingleType(element));
    }
    return isSupportedSingleType(value);
};
class Where {
    constructor(
    /** the where parameters to use */
    whereParams, 
    /** an existing bind param to be used, so the properties can be merged. If empty, a new one will be created and used */
    bindParam) {
        /** all the given options, so we can easily combine them into a new statement */
        this.rawParams = [];
        /**
         * an object with the key being the `identifier.property` and the value being the the bind param name which corresponds to it, and an operator to be used in the statement
         * this is needed for the following reasons:
         * 1) when generating the statement, those values are used
         * 2) the bind param names which are generated from this Where need to be differentiated from the actual keys of the bindParam, since this Where can only remove those
         */
        this.identifierPropertyData = [];
        /** adds a value to the bind param, while updating the usedBindParamNames field appropriately */
        this.addBindParamDataEntry = ({ identifier, property, operator, value, }) => {
            const bindParamName = this.bindParam.getUniqueNameAndAddWithLiteral(property, value);
            this.identifierPropertyData.push({
                identifier,
                property,
                bindParamName,
                operator,
            });
        };
        /** gets the statement by the params */
        this.getStatement = (
        /**
         * text is in the format "a.p1 = $v1 AND a.p2 = $v2"
         * object is in the format "{ a.p1 = $v1, a.p2 = $v2 }"
         */
        mode) => {
            const statementParts = [];
            const operatorForStatement = (operator) => {
                if (mode === 'object') {
                    if (operator !== 'eq') {
                        throw new Errors_1.NeogmaConstraintError('The only operator which is supported for object mode is "eq"', {
                            actual: {
                                mode,
                                operator,
                            },
                        });
                    }
                    // : is the only operator used in object mode
                    return ':';
                }
                const textMap = {
                    eq: '=',
                    in: 'IN',
                    _in: 'IN',
                    contains: 'CONTAINS',
                    gt: '>',
                    gte: '>=',
                    lt: '<',
                    lte: '<=',
                    ne: '<>',
                };
                // else, return the appropriate text-mode operator
                return textMap[operator];
            };
            /**
             * if false, it will be in the format: property operator param
             * if true, it will be in the format: param operator property
             */
            const isReverseOperator = (operator) => {
                if (operator === '_in') {
                    return true;
                }
                return false;
            };
            if (mode === 'text') {
                for (const bindParamData of this.identifierPropertyData) {
                    const { bindParamName } = bindParamData;
                    const name = bindParamName instanceof Literal_1.Literal
                        ? bindParamName.getValue()
                        : `$${bindParamName}`;
                    if (isReverseOperator(bindParamData.operator)) {
                        statementParts.push([
                            name,
                            operatorForStatement(bindParamData.operator),
                            `${bindParamData.identifier}.${bindParamData.property}`,
                        ].join(' '));
                    }
                    else {
                        statementParts.push([
                            `${bindParamData.identifier}.${bindParamData.property}`,
                            operatorForStatement(bindParamData.operator),
                            name,
                        ].join(' '));
                    }
                }
                return statementParts.join(' AND ');
            }
            if (mode === 'object') {
                for (const bindParamData of this.identifierPropertyData) {
                    const { bindParamName } = bindParamData;
                    const name = bindParamName instanceof Literal_1.Literal
                        ? bindParamName.getValue()
                        : `$${bindParamName}`;
                    statementParts.push([
                        bindParamData.property,
                        operatorForStatement(bindParamData.operator),
                        ` ${name}`,
                    ].join(''));
                }
                return `{ ${statementParts.join(', ')} }`;
            }
            throw new Errors_1.NeogmaConstraintError(`invalid mode ${mode}`);
        };
        this.bindParam = BindParam_1.BindParam.acquire(bindParam);
        this.addParams(whereParams);
        Object.setPrototypeOf(this, Where.prototype);
    }
    /** gets the BindParam used in this Where */
    getBindParam() {
        return this.bindParam;
    }
    /** gets the raw where parameters used to generate the final result */
    getRawParams() {
        return this.rawParams;
    }
    /** refreshes the statement and the bindParams by the given where params */
    addParams(
    /** the where parameters to use */
    whereParams) {
        // push the latest whereParams to the end of the array
        this.rawParams.push(whereParams);
        /* set the identifierPropertyData field by the rawParams */
        // merge all rawParams, for each identifier, into a single one. That way, the latest rawOption will dictate its properties if some previous ones have a common key
        const params = {};
        for (const rawParam of this.rawParams) {
            for (const [identifier, value] of Object.entries(rawParam)) {
                params[identifier] = Object.assign(Object.assign({}, params[identifier]), value);
            }
        }
        // remove all used bind param names from the bind param, since we're gonna set them again from scratch
        this.bindParam.remove(this.identifierPropertyData
            .map(({ bindParamName }) => bindParamName)
            .filter((v) => typeof v === 'string'));
        // reset identifierPropertyData as they've been removed from the bindParam
        this.identifierPropertyData = [];
        for (const nodeIdentifier in params) {
            for (const property in params[nodeIdentifier]) {
                const value = params[nodeIdentifier][property];
                if (isNeo4jSupportedTypes(value)) {
                    this.addBindParamDataEntry({
                        identifier: nodeIdentifier,
                        property,
                        value,
                        operator: 'eq',
                    });
                }
                else if (value !== null && value !== undefined) {
                    for (const operator of operators) {
                        if (isOperator[operator](value)) {
                            this.addBindParamDataEntry({
                                identifier: nodeIdentifier,
                                property,
                                value: value[exports.Op[operator]],
                                operator,
                            });
                            break;
                        }
                    }
                }
            }
        }
        return this;
    }
    /** returns a Where object if params is specified, else returns null */
    static acquire(params, bindParam) {
        if (!params) {
            return null;
        }
        if (params instanceof Where) {
            return params;
        }
        return new Where(params, bindParam);
    }
    /**
     * if the value is not an array, it gets returned as is. If it's an array, a "[Op.in]" object is returned for that value
     */
    static ensureIn(value) {
        return value instanceof Array
            ? {
                [exports.Op.in]: value,
            }
            : value;
    }
}
exports.Where = Where;
//# sourceMappingURL=Where.js.map